from __future__ import division
import numpy as np
import soundfile as sf
from . import timbral_util


def boominess_calculate(loudspec):
    """
      Calculates the Booming Index as described by Hatano, S., and Hashimoto, T. "Booming index as a measure for
      evaluating booming sensation", The 29th International congress and Exhibition on Noise Control Engineering, 2000.
    """

    # loudspec from the loudness_1991 code results in values from 0.1 to 24 Bark in 0.1 steps
    # I need to calculate a new weighting function for this
    z = np.arange(0.1, 24.05, 0.1)
    f = 600 * np.sinh(z / 6.0)  # convert these bark values to frequency
    FR = [25, 31.5, 40, 50, 63, 80, 100, 125, 160, 200, 250, 315, 400, 500, 630, 800, 1000, 1250, 1600, 2000, 2500,
          3150, 4000, 5000, 6300, 8000, 10000, 12500] # get the centre frequencies of 3rd octave bands

    # now I need to convert f onto the FR scale
    logFR = np.log10(FR)
    FR_step = logFR[1] - logFR[0]  # get the step size on the log scale
    FR_min = logFR[0]  # get the minimum value of the logFR

    logf = np.log10(f)  # get the log version of estimated frequencies
    # estimate the indexes of the bark scale on the 3rd octave scale
    estimated_index = ((logf - FR_min) / float(FR_step)) + 1

    # weighting function based from the estimated indexes
    Weighting_function = 2.13 * np.exp(-0.151 * estimated_index)
    Weighting_function = 4.43 * np.exp(-0.151 * estimated_index)

    # change the LF indexes to roll off
    Weighting_function[0] = 0.5  # this value is estimated
    Weighting_function[1] = 1.05
    Weighting_function[2] = 1.10
    Weighting_function[3] = 1.18


    # identify index where frequency is less than 280Hz
    below_280_idx = np.where(f >= 280)[0][0]

    I = loudspec * Weighting_function
    loudness = np.sum(loudspec)
    Ll = np.sum(loudspec[:below_280_idx])

    Bandsum = timbral_util.log_sum(I)
    BoomingIndex = Bandsum * (Ll / loudness)


    '''
      Trying a new weighting function
    '''
    wf = 4.43 * np.exp(-0.151 * np.linspace(0, 24, 240))
    lf_decay_inc = 0.1
    for i in range(20, -1, -1):
        wf[i] -= lf_decay_inc
        lf_decay_inc += 0.1

    new_I = np.sum(loudspec * wf)
    low_loudness = np.sum(loudspec[:40])
    low_ratio = low_loudness / np.sum(loudspec)

    # plt.plot(wf)


    return BoomingIndex, new_I, low_loudness, low_ratio


def timbral_booming(fname, dev_output=False, phase_correction=False):
    """
     This is an implementation of the hasimoto booming index feature.
     There are a few fudge factors with the code to convert between the internal representation of the sound using the
     same loudness calculation as the sharpness code.  The equation for calculating the booming index is not
     specifically quoted anywhere so I've done the best i can with the code that was presented.

     Shin, SH, Ih, JG, Hashimoto, T., and Hatano, S.: "Sound quality evaluation of the booming sensation for passenger
      cars", Applied Acoustics, Vol. 70, 2009.

     Hatano, S., and Hashimoto, T. "Booming index as a measure for
      evaluating booming sensation", The 29th International congress and Exhibition on Noise Control Engineering, 2000.

     This function calculates the apparent Boominess of an audio file.

     Version 0.2

    Required parameter
    :param fname:                   Audio filename to be analysed, including full file path and extension.

    Optional parameters
    :param dev_output:              Bool, when False return the warmth, when True return all extracted features (currently none)
    :param phase_correction:        If the inter-channel phase should be estimated when performing a mono sum.
                                    Defaults to False.

    :return                         Apparent boominess of the audio file.

    Copyright 2018 Andy Pearce, Institute of Sound Recording, University of Surrey, UK.

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.


   http://www.conforg.fr/internoise2000/cdrom/data/articles/000233.pdf

    """

    # use pysoundfile to read audio
    audio_samples, fs = sf.read(fname, always_2d=False)
    audio_samples = timbral_util.channel_reduction(audio_samples, phase_correction=phase_correction)

    # normalisation disabled because the values are dependent on the actual loudness odf the audio, which may get
    # screwed up during normalisation
    # normalise audio
    # audio_samples /= max(abs(audio_samples))




    #  i want to caluclate the specific loudness over a frame by frame basis.

    # pad with zeros to make division possible
    window_length = 4096
    remainder = np.mod(len(audio_samples), window_length)
    audio_samples = np.pad(audio_samples, (0, window_length-remainder), 'constant', constant_values=0.0)
    windowed_samples = np.reshape(audio_samples, (int(len(audio_samples) / window_length), window_length))


    windowed_booming = []
    windowed_rms = []
    windowed_loudness = []
    windowed_newI = []
    windowed_low_loudness = []
    windowed_low_ratio = []
    for i in range(windowed_samples.shape[0]):
        if i == 28:
            x = 1
        samples = windowed_samples[i, :]

        windowed_rms.append(np.sqrt(np.mean(samples * samples)))

        # calculate the specific loudness
        N_entire, N_single = timbral_util.specific_loudness(samples, Pref=100.0, fs=fs, Mod=0)

        windowed_loudness.append(N_entire)

        # calculate the booming index
        if N_entire > 0:
            # boom = boominess_calculate(N_single)
            BoomingIndex, new_I, low_loudness, low_ratio = boominess_calculate(N_single)
        else:
            BoomingIndex = 0
            new_I = 0
            low_loudness = 0
            low_ratio = 0
            boom = 0

        windowed_booming.append(BoomingIndex)
        windowed_newI.append(new_I)
        windowed_low_loudness.append(low_loudness)
        windowed_low_ratio.append(low_ratio)

    '''  new bark level stuff '''
    ll, w_ll = timbral_util.weighted_bark_level(audio_samples, fs, 0, 70)
    hl, w_hl = timbral_util.weighted_bark_level(audio_samples, fs, 70, 240)

    # convert to numpy arrays for fancy indexing
    windowed_booming = np.array(windowed_booming)
    windowed_rms = np.array(windowed_rms)

    boom = np.mean(windowed_booming)
    rms_boom = np.average(windowed_booming, weights=windowed_rms)
    rms_boom2 = np.average(windowed_booming, weights=(windowed_rms * windowed_rms))
    loudness_boom = np.average(windowed_booming, weights=windowed_loudness)


    newI = np.average(windowed_newI, weights=windowed_rms)
    low_loudness = np.average(windowed_low_loudness, weights=windowed_rms)
    low_ratio = np.average(windowed_low_ratio, weights=windowed_rms)





    if dev_output:
        return [boom, rms_boom, loudness_boom, newI, low_loudness, low_ratio, ll, w_ll, hl, w_hl, rms_boom2]
    else:
        return rms_boom
