import numpy as np
import soundfile as sf
import timbral_util
import scipy.stats
from scipy.signal import spectrogram
from Cython_test import cython_smoothing


def calcFrequencyScale(spectrum_length, fs, blockSize):
    """ calculate frequency scale """
    freq = np.array(range(0, spectrum_length)) * fs / float(blockSize)
    return freq


def calcLowFrequencyLimit(fls, noct, max_idx):
    """ Calculates an array containing the indexes of lower frequency
    bounds for the spectral smoothing. """
    # floats required due to integer division in Python 2.7
    f_lower = fls[0:max_idx] / (2.0 ** (1 / (2.0 * noct)))
    step_size = fls[1] - fls[0]
    approx_idx = f_lower / (1.0 * step_size)
    f_lower = np.round(approx_idx).astype(int)
    return f_lower


def calcUpperFrequencyLimit(fls, noct, max_idx):
    """ calculate an array containing the indexes of upper frequency
    bounds for the spectral smoothing. """
    # floats required due to integer division in Python 2.7
    f_upper = fls[0:max_idx] * (2.0 ** (1.0 / (2.0 * noct)))
    step_size = fls[1] - fls[0]
    approx_idx = f_upper / float(step_size)
    f_upper = np.round(approx_idx).astype(int)
    return f_upper


def calcMaxIDX(fls, noct):
    """ calculates the max index of the frequency scale with respect
    to the level of smoothing applied.  This ensures that only
    frequencies below the Nyquist limit are used for smoothing """
    freq_l = fls[-1] / (2.0 ** (1 / (2.0 * noct)))
    max_idx = np.array(abs(fls - freq_l)).argmin()
    return max_idx


def calcMinIDX(fls, minFreq):
    """ calculates the index of the minimum frequency used for
    smoothing.  This ensures that frequencies below the lower
    limit are not considered.  """
    min_idx = np.argmax(fls >= minFreq)
    return min_idx


def oldSpectralSmoothing(spectrum, f_upper, f_lower):
    """ Performs the spectral smoothing on a step-by-step basis.
    For each valid frequency bin, the magnitude is averaged over
    1/Noct below and above the centre frequency.
    """
    smoothed_array = np.zeros(len(f_upper))
    for i in range(len(smoothed_array)):
        # if statement is required since clever indexing isn't that clever.
        if f_upper[i] == f_lower[i]:
            smoothed_array[i] = spectrum[f_lower[i]]
        else:
            smoothed_array[i] = np.mean(spectrum[f_lower[i]:(f_upper[i])])
            # smooth_values = spectrum[f_lower[i]:(f_upper[i] + 1)]
            # smoothed_array[i] = np.mean(smooth_values)
    return smoothed_array

def spectralSmoothing(spectrum, f_upper, f_lower):
    """ Performs the spectral smoothing on an entire spectrogram.
    For each valid frequency bin, the magnitude is averaged over
    1/Noct below and above the centre frequency.
    """
    smoothed_spec = np.zeros((len(f_upper), spectrum.shape[1]))
    # smoothed_array = np.zeros(())
    for i in range(len(f_upper)):
        upper_lim = f_upper[i]
        lower_lim = f_lower[i]

        if upper_lim == lower_lim:
            smoothed_spec[i,:] = spectrum[lower_lim, :]
        else:
            hold_spec = spectrum[lower_lim:upper_lim+1, :]
            mean_spec = np.mean(hold_spec, axis=0)
            smoothed_spec[i, :] = mean_spec

    return smoothed_spec


def calcCrossoverIDX(fls, crossover_freq):
    """ Calculates the index of the crossover frequency """
    crossover_idx = np.argmax(fls >= crossover_freq)
    return crossover_idx

# def evaluate_smoothing_speed(current_spectrum, f_upper, f_lower):
#     """ Measure the time that the queries need to return their results """
#     from timeit import Timer
#
#     t = Timer("spectralSmoothing(current_spectrum, f_upper, f_lower)",
#               "from timbral_models.old_timbral_brightness import *")
#
#     sf = 'Execution time: {} ms'
#     print(sf.format(t.timeit(number=20)))


def timbral_brightness(fname, dev_output=False, threshold=0, crossover=3000, stepSize=1024, blockSize=2048, threshold_db=-60, n_oct=2,
                       minFreq=20, phase_correction=False, smoothing_method='python', maxFreq=0):
    """ The main process block where computations are completed with VAMP.
     Data is accepted in the time domain, where a Hamming window is applied,
     and the FFT of this data taken.  The magnitude spectrum is then smoothed
     according to Noct (set to half-octave band smoothing by default),
     before calculating the spectral centroid above the crossover frequency,
     and the ratio of energy above the crossover frequency to all energy.
    """
    # use pysoundfile instead
    audio_samples, fs = sf.read(fname, always_2d=False)

    # get sum all channels to mono
    num_channels = np.shape(audio_samples)
    if len(num_channels) > 1:
        # crudely check for out of phase signals
        if phase_correction:
            r, pval = scipy.stats.pearsonr(audio_samples[:, 0], audio_samples[:, 1])
            if r < -0.5:
                audio_samples = audio_samples[:, 0]  # [:,1] *= -1.0
            else:
                audio_samples = np.sum(audio_samples, axis=1)
        else:
            audio_samples = np.sum(audio_samples, axis=1)

    '''
      Filter audio
    '''
    # highpass audio at minimum frequency
    audio_samples = timbral_util.filter_audio_highpass(audio_samples, crossover=minFreq, fs=fs)
    audio_samples = timbral_util.filter_audio_highpass(audio_samples, crossover=minFreq, fs=fs)
    audio_samples = timbral_util.filter_audio_highpass(audio_samples, crossover=minFreq, fs=fs)
    # nfft = blockSize
    # hop_size = 3 * nfft / 4
    # freq, time, spec1 = spectrogram(audio_samples, fs, 'hamming', nfft, hop_size, nfft, 'constant', True, 'spectrum')

    # lowpass audio at maximumfrequency
    if maxFreq > 0:
        audio_samples = timbral_util.filter_audio_lowpass(audio_samples, crossover=maxFreq, fs=fs)
        audio_samples = timbral_util.filter_audio_lowpass(audio_samples, crossover=maxFreq, fs=fs)
        audio_samples = timbral_util.filter_audio_lowpass(audio_samples, crossover=maxFreq, fs=fs)
    # freq, time, spec2 = spectrogram(audio_samples, fs, 'hamming', nfft, hop_size, nfft, 'constant', True, 'spectrum')

    # get highpass audio at crossover
    highpass_audio = timbral_util.filter_audio_highpass(audio_samples, crossover, fs)
    highpass_audio = timbral_util.filter_audio_highpass(highpass_audio, crossover, fs)
    highpass_audio = timbral_util.filter_audio_highpass(highpass_audio, crossover, fs)

    # lowpass audio at crossover
    lowpass_audio = timbral_util.filter_audio_lowpass(audio_samples, crossover, fs)
    lowpass_audio = timbral_util.filter_audio_lowpass(lowpass_audio, crossover, fs)
    lowpass_audio = timbral_util.filter_audio_lowpass(lowpass_audio, crossover, fs)

    '''
     Get spectrograms and normalise
    '''
    # normalise audio
    lowpass_audio *= (1.0 / max(abs(audio_samples)))
    highpass_audio *= (1.0 / max(abs(audio_samples)))
    audio_samples *= (1.0 / max(abs(audio_samples)))

    # set FFT parameters
    nfft = blockSize
    hop_size = 3 * nfft / 4
    # get spectrogram
    freq, time, spec = spectrogram(audio_samples, fs, 'hamming', nfft, hop_size, nfft, 'constant', True, 'spectrum')
    lp_freq, lp_time, lp_spec = spectrogram(lowpass_audio, fs, 'hamming', nfft, hop_size, nfft, 'constant', True,
                                            'spectrum')
    hp_freq, hp_time, hp_spec = spectrogram(highpass_audio, fs, 'hamming', nfft, hop_size, nfft, 'constant', True,
                                            'spectrum')

    # get variables for smoothing
    if n_oct > 0:
        maxIDX = calcMaxIDX(freq, n_oct)
        f_upper = calcUpperFrequencyLimit(freq, n_oct, maxIDX)
        f_lower = calcLowFrequencyLimit(freq, n_oct, maxIDX)

        if smoothing_method == 'python':
            # original smoothing
            spec = spectralSmoothing(spec, f_upper, f_lower)
            lp_spec = spectralSmoothing(lp_spec, f_upper, f_lower)
            hp_spec = spectralSmoothing(hp_spec, f_upper, f_lower)
        elif smoothing_method == 'cython':
            # cython smoothing
            spec = cython_smoothing(spec, f_upper, f_lower)
            lp_spec = cython_smoothing(lp_spec, f_upper, f_lower)
            hp_spec = cython_smoothing(hp_spec, f_upper, f_lower)


    if threshold_db == 0:
        threshold = 0
        hp_threshold = 0
    else:
        max_power = max(np.sum(spec, axis=1))
        # min_power = min(np.sum(spec, axis=1))
        threshold = max_power * timbral_util.db2mag(threshold_db)
        hp_max_power = max(np.sum(hp_spec, axis=1))
        # hp_min_power = min(np.sum(hp_spec, axis=1))
        hp_threshold = hp_max_power * timbral_util.db2mag(threshold_db)

    all_ratio1 = []
    all_ratio2 = []
    all_centroid =[]
    all_hp_centroid = []
    all_tpower = []
    all_hp_tpower = []

    for idx in range(len(time)):
        # if idx > 1230:
        #     x = 1
        # get overall spectrum of time frame
        current_spectrum = spec[:, idx]
        # calculate time window power
        tpower = np.sum(current_spectrum)

        # estimate if time segment contains audio energy or just noise
        if tpower > threshold:
            lp_spectrum = lp_spec[:, idx]
            hp_spectrum = hp_spec[:, idx]

            tpower = np.sum(current_spectrum)
            lp_tpower = np.sum(lp_spectrum)
            hp_tpower = np.sum(hp_spectrum)

            all_tpower.append(tpower)
            # all_hp_tpower.append(hp_tpower)

            all_centroid.append(np.sum(current_spectrum * freq[:len(current_spectrum)]) / tpower)
            # all_hp_centroid.append(np.sum(hp_spectrum * hp_freq[:len(hp_spectrum)]) / float(np.sum(hp_spectrum)))

            all_ratio1.append(hp_tpower / lp_tpower)
            all_ratio2.append(hp_tpower / tpower)


    '''
      high pass threshold version
    '''
    for idx in range(len(time)):
        # get overall spectrum of time frame
        hp_spectrum = hp_spec[:, idx]
        # calculate time window power
        hp_tpower = np.sum(hp_spectrum)

        # estimate if time segment contains audio energy or just noise
        if hp_tpower > hp_threshold:
            lp_spectrum = lp_spec[:, idx]
            current_spectrum = spec[:, idx]

            tpower = np.sum(current_spectrum)
            lp_tpower = np.sum(lp_spectrum)
            hp_tpower = np.sum(hp_spectrum)

            # all_tpower.append(tpower)
            all_hp_tpower.append(hp_tpower)

            # all_centroid.append(np.sum(current_spectrum * freq[:len(current_spectrum)]) / tpower)
            all_hp_centroid.append(np.sum(hp_spectrum * hp_freq[:len(hp_spectrum)]) / float(np.sum(hp_spectrum)))

            # all_ratio1.append(hp_tpower / lp_tpower)
            # all_ratio2.append(hp_tpower / tpower)


    mean_ratio1 = np.mean(all_ratio1)
    mean_ratio2 = np.mean(all_ratio2)
    mean_centroid = np.mean(all_centroid)
    mean_hp_centroid = np.mean(all_hp_centroid)

    # print mean_hp_centroid

    weighted_mean_ratio1 = np.average(all_ratio1, weights=all_tpower)
    weighted_mean_ratio2 = np.average(all_ratio2, weights=all_tpower)
    weighted_mean_centroid = np.average(all_centroid, weights=all_tpower)
    weighted_mean_hp_centroid = np.average(all_hp_centroid, weights=all_hp_tpower)



    if dev_output:
        # return ratio, mean_centroid, np.log10(ratio), np.log10(mean_centroid), ratio*mean_centroid, np.log10(ratio)*np.log10(mean_centroid)
        return   mean_ratio1, mean_ratio2, mean_centroid, mean_hp_centroid, weighted_mean_ratio1, \
                 weighted_mean_ratio2, weighted_mean_centroid, weighted_mean_hp_centroid, \
                 np.log10(mean_ratio1), np.log10(mean_ratio2), np.log10(mean_centroid), np.log10(mean_hp_centroid), \
                 np.log10(weighted_mean_ratio1), np.log10(weighted_mean_ratio2), np.log10(weighted_mean_centroid), \
                 np.log10(weighted_mean_hp_centroid)
    else:
        return bright
