import numpy as np
import soundfile as sf
from scipy.signal import spectrogram
import pyfilterbank
# import essentia
# import essentia.standard as es
import timbral_util
import scipy.stats


def db2mag(dB):
    mag = 10 ** (dB / 20.0)
    return mag


def timbral_depth(fname, dev_output=False, threshold_db=-60, low_frequency_limit=30, crossover_frequency=200,
                  phase_correction=False, weighting='None', db_decay_threshold=-40):
    """
     This function calculates the apparent Depth of an audio file.
     Currently the function only works on the left channel of multi-channel audio files.

     The function calculates 3 parameters: the spectral centroid of low frequencies, ratio of low-frequency
     to all energy, and the roll-on frequency.

      Spectral centroid:    The spectral centroid is calculated for frequencies between 30 and 200 Hz.  A window
                            size of 4096 is used to obtain higher low-frequency resolution, and a 1024 samples hop
                            size to maintain time resolution.

      Ratio:                Calculates the ratio fo magnitude between 30 to 200 Hz compared with magnitude between
                            30 Hz and Nyquist.

      Roll-on frequency:    This calculates the frequency where 95% of energy lies above this frequency.

      Depth:                The depth is calculated by applying the coefficients obtained from a linear regression
                            model, which are hardcoded into this function.

      :param fname: Audio filename to be analysed, including full file path and extension.
      :return: returns the depth as a float representing the relative depth.
    """
    # use pysoundfile to read audio
    audio_samples, fs = sf.read(fname, always_2d=False)

    audio_samples = timbral_util.channel_reduction(audio_samples)


    '''
      New feature
    '''
    lowband_level, weighed_lowband_level = timbral_util.weighted_bark_level(audio_samples, fs, low_bark_band=10,
                                                                            upper_bark_band=60)

    midlowband_level, weighed_midlowband_level = timbral_util.weighted_bark_level(audio_samples, fs, low_bark_band=10,
                                                                            upper_bark_band=120)

    # overall levels
    # N_entire, N_single = specific_loudness(audio_samples, Pref=100.0, fs=fs, Mod=0)

    # take left channel
    num_channels = np.shape(audio_samples)
    if len(num_channels) > 1:
        # take just the left channel
        audio_samples = audio_samples[:, 0]

    # highpass audio - running 3 times for 18dB per octave
    audio_samples = timbral_util.filter_audio_highpass(audio_samples, crossover=low_frequency_limit, fs=fs)
    audio_samples = timbral_util.filter_audio_highpass(audio_samples, crossover=low_frequency_limit, fs=fs)
    audio_samples = timbral_util.filter_audio_highpass(audio_samples, crossover=low_frequency_limit, fs=fs)

    if weighting != 'None':
        audio_samples = pyfilterbank.splweighting.weight_signal(audio_samples, fs, weighting)

    # running 3 times to get -18dB per octave rolloff, greater than second order filters are unstable in python
    lowpass_audio_samples = timbral_util.filter_audio_lowpass(audio_samples, crossover=crossover_frequency, fs=fs)
    lowpass_audio_samples = timbral_util.filter_audio_lowpass(lowpass_audio_samples, crossover=crossover_frequency, fs=fs)
    lowpass_audio_samples = timbral_util.filter_audio_lowpass(lowpass_audio_samples, crossover=crossover_frequency, fs=fs)

    highpass_audio_samples = timbral_util.filter_audio_highpass(audio_samples, crossover=crossover_frequency, fs=fs)
    highpass_audio_samples = timbral_util.filter_audio_highpass(audio_samples, crossover=crossover_frequency, fs=fs)
    highpass_audio_samples = timbral_util.filter_audio_highpass(audio_samples, crossover=crossover_frequency, fs=fs)

    if weighting != 'None':
        audio_samples = pyfilterbank.splweighting.weight_signal(audio_samples, fs, weighting)
        lowpass_audio_samples = pyfilterbank.splweighting.weight_signal(lowpass_audio_samples, fs, weighting)
        highpass_audio_samples = pyfilterbank.splweighting.weight_signal(highpass_audio_samples, fs, weighting)


    '''
      pre normalising values
    '''
    # set FFT parameters
    nfft = 4096
    hop_size = 3 * nfft / 4
    # get spectrogram
    freq, time, spec = spectrogram(audio_samples, fs, 'hamming', nfft,
                                   hop_size, nfft, 'constant', True, 'spectrum')
    lp_freq, lp_time, lp_spec = spectrogram(lowpass_audio_samples, fs, 'hamming', nfft,
                                            hop_size, nfft, 'constant', True, 'spectrum')

    '''
      get the overall spectral metrics
    '''
    # get overall spectral metrics
    all_spectrum = np.sum(spec, axis=1)

    log_spectrum = 20 * np.log10(all_spectrum)
    log_spectrum += (0 - max(log_spectrum))
    shifted_log_spectrum = log_spectrum + abs(min(log_spectrum))
    # get spectral metrics
    SC = np.sum(freq * all_spectrum) / float(np.sum(all_spectrum))
    flogSC = 10 ** (np.sum(np.log10(freq[1:]) * all_spectrum[1:]) / float(np.sum(all_spectrum[1:])))
    logSC = np.sum(freq * shifted_log_spectrum) / float(np.sum(shifted_log_spectrum))
    floglogSC = 10 ** (np.sum(np.log10(freq[1:]) * shifted_log_spectrum[1:]) / float(np.sum(shifted_log_spectrum[1:])))
    freq_count = 1
    log_tpower_thresh = 0.05 * np.sum(shifted_log_spectrum)
    lin_tpower_thresh = 0.05 * np.sum(all_spectrum)
    log_tpower = shifted_log_spectrum[0]
    lin_tpower = all_spectrum[0]

    while log_tpower < log_tpower_thresh:
        freq_count += 1
        log_tpower = np.sum(shifted_log_spectrum[:freq_count])

    log_rollon = float(freq[freq_count])

    freq_count = 1
    while lin_tpower < lin_tpower_thresh:
        freq_count += 1
        lin_tpower = np.sum(all_spectrum[:freq_count])

    lin_rollon = float(freq[freq_count])


    '''
      My f0 estimation with peak picking
    '''
    norm_spec = (all_spectrum - np.min(all_spectrum)) / (np.max(all_spectrum) - np.min(all_spectrum))
    cthr = 0.01

    # array, freq = 0, cthr = 0.2, unprocessed_array = False, fs = 44100

    peak_idx, peak_value, peak_freq = timbral_util.detect_peaks(norm_spec, cthr=cthr, fs=fs)
    # peak_idx, peak_value, peak_freq = timbral_util.detect_peaks(norm_spec, cthr, norm_spec, freq)
    pitch_estimate = np.log10(min(peak_freq)) if peak_freq[0] > 0 else 0




    '''
      get frame based spectral metrics
    '''

    # get the index for the low-frequency limit, 30 Hz
    # low_limit_idx = np.argmax(freq >= 30)

    # get the index for the cross-over frequency, 200 Hz
    # crossover_idx = np.argmax(freq >= 200)

    # set threshold for ignoring time segments with no noise, selected from the training stimuli
    threshold = 0.005 ** 2
    threshold = db2mag(threshold_db)  # 0.005 ** 2

    # define arrays for storing metrics
    all_lower_centroid = []
    all_thx_lower_centroid = []
    all_lower_ratio = []
    all_thx_lower_ratio = []
    all_rollon = []
    all_thx_rollon = []
    all_lowpower = []
    all_thx_lowpower = []
    all_tpower = []

    # get metrics for each time segment of the spectrogram
    for idx in range(1, len(time)):
        # current_spectrum = spec[low_limit_idx:, idx]
        # current_lp_spectrum = lp_spec[low_limit_idx:, idx]
        current_spectrum = spec[:, idx]
        current_lp_spectrum = lp_spec[:, idx]
        tpower = np.sum(current_spectrum)
        all_tpower.append(tpower)

        # estimate if time segment contains audio energy or just noise
        if tpower > threshold:
            # calculate the spectral centroid and ratio
            # lower_spectrum = spec[low_limit_idx:crossover_idx, idx]
            # lower_fls = freq[low_limit_idx:crossover_idx]
            # lower_power = np.sum(lower_spectrum)
            lower_power = np.sum(current_lp_spectrum)
            all_thx_lowpower.append(lower_power)
            all_lowpower.append(lower_power)

            lower_centroid = np.sum(current_lp_spectrum * lp_freq) / float(lower_power)
            lower_ratio = lower_power / float(tpower)

            # calculate the roll-on frequency
            cumulative_spectral_power = current_spectrum[0]
            counter = 0
            rollon_threshold = tpower * 0.05
            while cumulative_spectral_power < rollon_threshold:
                counter += 1
                cumulative_spectral_power = np.sum(current_spectrum[:counter])

            # rollon = freq[counter + low_limit_idx]
            rollon = freq[counter]

            all_lower_centroid.append(lower_centroid)
            all_thx_lower_centroid.append(lower_centroid)
            all_lower_ratio.append(lower_ratio)
            all_thx_lower_ratio.append(lower_ratio)
            all_rollon.append(rollon)
            all_thx_rollon.append(rollon)

            # print all_lower_centroid

        else:
            all_lower_centroid.append(0)
            all_lower_ratio.append(0)
            all_rollon.append(0)
            all_lowpower.append(np.sum(current_lp_spectrum))

    '''
      get normalised values
    '''
    # normalise audio
    lowpass_audio_samples *= (1.0 / max(abs(audio_samples)))
    audio_samples *= (1.0 / max(abs(audio_samples)))

    # set FFT parameters
    nfft = 4096
    hop_size = 3 * nfft / 4
    # get spectrogram
    freq, time, spec = spectrogram(audio_samples, fs, 'hamming', nfft,
                                   hop_size, nfft, 'constant', True, 'spectrum')
    lp_freq, lp_time, lp_spec = spectrogram(lowpass_audio_samples, fs, 'hamming', nfft,
                                   hop_size, nfft, 'constant', True, 'spectrum')

    # set threshold for ignoring time segments with no noise, selected from the training stimuli
    threshold = db2mag(threshold_db) # 0.005 ** 2


    # define arrays for storing metrics
    all_normalised_lower_centroid = []
    all_normalised_thx_lower_centroid = []
    all_normalised_lower_ratio = []
    all_normalised_thx_lower_ratio = []
    all_normalised_rollon = []
    all_normalised_thx_rollon = []
    all_normalised_lowpower = []
    all_normalised_thx_lowpower = []
    all_normalised_tpower = []
    all_centroid = []


    # get metrics for each time segment of the spectrogram
    for idx in range(1, len(time)):
        # current_spectrum = spec[low_limit_idx:, idx]
        # tpower = np.sum(current_spectrum)
        # all_normalised_tpower.append(tpower)
        current_spectrum = spec[:, idx]
        tpower = np.sum(current_spectrum)
        all_normalised_tpower.append(tpower)

        # estimate if time segment contains audio energy or just noise
        if tpower > threshold:
            # calculate the spectral centroid and ratio
            # all_freq = freq[low_limit_idx:]
            centroid = np.sum(current_spectrum * freq) / tpower
            all_centroid.append(centroid)
            # lower_spectrum = spec[low_limit_idx:crossover_idx, idx]
            # lower_fls = freq[low_limit_idx:crossover_idx]
            lower_spectrum = lp_spec[:, idx]
            lower_fls = lp_freq
            lower_power = np.sum(lower_spectrum)
            all_normalised_thx_lowpower.append(lower_power)
            all_normalised_lowpower.append(lower_power)


            lower_centroid = np.sum(lower_spectrum * lower_fls) / float(lower_power)
            lower_ratio = lower_power / float(tpower)

            # calculate the roll-on frequency
            cumulative_spectral_power = current_spectrum[0]
            counter = 0
            rollon_threshold = tpower * 0.05
            while cumulative_spectral_power < rollon_threshold:
                counter += 1
                cumulative_spectral_power = np.sum(current_spectrum[:counter])

            # rollon = freq[counter + low_limit_idx]
            rollon = freq[counter]

            all_normalised_lower_centroid.append(lower_centroid)
            all_normalised_thx_lower_centroid.append(lower_centroid)
            all_normalised_lower_ratio.append(lower_ratio)
            all_normalised_thx_lower_ratio.append(lower_ratio)
            all_normalised_rollon.append(rollon)
            all_normalised_thx_rollon.append(rollon)

            # print all_lower_centroid

        else:
            all_normalised_lower_centroid.append(0)
            all_normalised_lower_ratio.append(0)
            all_normalised_rollon.append(0)
            all_normalised_thx_lowpower.append(0)
            all_centroid.append(0)

            # lower_spectrum = spec[low_limit_idx:crossover_idx, idx]
            lower_spectrum = lp_spec[:, idx]
            lower_power = np.sum(lower_spectrum)
            all_normalised_lowpower.append(lower_power)











    '''
      pitch estimation
    '''
    # pitch_estimator = es.PitchYin(sampleRate=fs)
    # es_audio = essentia.array(audio_samples)
    # es_rms_hold = []
    # pitch_hold = []
    # pitchCI_hold = []
    # for frame in es.FrameGenerator(es_audio, frameSize=2048, hopSize=512, startFromZero=True):
    #     es_rms_hold.append(np.sqrt(np.mean(frame * frame)))
    #     pitch, pitchCI = pitch_estimator(essentia.array(frame))
    #     pitch_hold.append(pitch)
    #     pitchCI_hold.append(pitchCI)
    #
    #
    # # import matplotlib.pyplot as plt
    # # plt.plot(es_rms_hold)
    # #
    # weighted_pitch = np.average(pitch_hold, weights=es_rms_hold)
















    '''
            1) segment audio file
            2) identify decay start
            3) estimate time-domain decay time
            4) estimate LF/HF decay time
        '''
    # get essentia StrongDecay
    # es_array = essentia.array(audio_samples)
    # lp_es_array = essentia.array(lowpass_audio_samples)
    # strong_decay = float(es.StrongDecay()(es_array))
    # lp_strong_decay = float(es.StrongDecay()(lp_es_array))

    # get essentia effective duration
    effective_duration = []
    lp_effective_duration = []
    all_my_duration = []
    lp_all_my_duration = []
    hp_all_my_duration = []


    # get envelpe of signal
    envelope = timbral_util.sample_and_hold_envelope_calculation(audio_samples, fs)
    onsets = timbral_util.calculate_onsets(audio_samples, envelope, fs)

    # initialise essentia function
    # es_effective_durection = es.EffectiveDuration(sampleRate=fs)
    # calculate essentia function per event
    for idx, onset in enumerate(onsets):
        if onset == onsets[-1]:
            segment = audio_samples[onset:]
            lp_segment = lowpass_audio_samples[onset:]
        else:
            segment = audio_samples[onset:onsets[idx + 1]]
            lp_segment = lowpass_audio_samples[onset:onsets[idx + 1]]

        # effective_duration.append(es_effective_durection(essentia.array(segment)))
        # lp_effective_duration.append(es_effective_durection(essentia.array(lp_segment)))

    # calculate my version of the decay time / effective time
    rms_step_size = 256
    rms_envelope = timbral_util.calculate_rms_enveope(audio_samples, step_size=rms_step_size)
    lowpass_rms_envelope = timbral_util.calculate_rms_enveope(lowpass_audio_samples, step_size=rms_step_size)
    highpass_rms_envelope = timbral_util.calculate_rms_enveope(highpass_audio_samples, step_size=rms_step_size)

    # convert to magnitude
    decay_threshold = timbral_util.db2mag(db_decay_threshold)
    time_convert = fs / float(rms_step_size)
    # rescale onsets to rms stepsize
    onsets = (np.array(onsets) / float(rms_step_size)).astype('int')

    for idx, onset in enumerate(onsets):
        if onset == onsets[-1]:
            segment = rms_envelope[onset:]
            lp_segment = lowpass_rms_envelope[onset:]
            hp_segment = highpass_rms_envelope[onset:]
        else:
            segment = rms_envelope[onset:onsets[idx + 1]]
            lp_segment = lowpass_rms_envelope[onset:onsets[idx + 1]]
            hp_segment = highpass_rms_envelope[onset:onsets[idx + 1]]

        # calculate my duration estimation
        max_idx = np.argmax(segment)
        post_max_segment = segment[max_idx:]
        lp_post_max_segment = lp_segment[max_idx:]
        hp_post_max_segment = hp_segment[max_idx:]

        if min(post_max_segment) >= decay_threshold:
            my_duration = len(post_max_segment) / time_convert
        else:
            my_duration = np.where(post_max_segment < decay_threshold)[0][0] / time_convert
        all_my_duration.append(my_duration)  # correct for RMS step size

        if min(lp_post_max_segment) >= decay_threshold:
            lp_my_duration = len(lp_post_max_segment) / float(fs)
        else:
            lp_my_duration = np.where(lp_post_max_segment < decay_threshold)[0][0] / float(fs)
        lp_all_my_duration.append(lp_my_duration)

        if min(hp_post_max_segment) >= decay_threshold:
            hp_my_duration = len(hp_post_max_segment) / float(fs)
        else:
            hp_my_duration = np.where(hp_post_max_segment < decay_threshold)[0][0] / float(fs)
        hp_all_my_duration.append(hp_my_duration)

    mean_effective_duration = np.mean(effective_duration)
    mean_my_duration = np.mean(all_my_duration)
    lp_mean_effective_duration = np.mean(lp_effective_duration)
    lp_mean_my_duration = np.mean(lp_all_my_duration)

    duration_ratio = 0 #np.array(lp_all_my_duration) / np.array(hp_all_my_duration)
    normalised_duration_ratio = np.array(duration_ratio) / np.array(all_my_duration)

    mean_duration_ratio = np.mean(duration_ratio)
    mean_normalised_duration_ratio = np.mean(normalised_duration_ratio)






    # get mean values
    mean_lower_centroid = np.mean(all_lower_centroid)
    mean_lower_ratio = np.mean(all_lower_ratio)
    mean_rollon = np.mean(all_rollon)
    mean_lowpower = np.mean(all_lowpower)

    # check for empty arrays
    mean_thx_lower_centroid = np.mean(all_thx_lower_centroid) if all_thx_lower_centroid else 0
    mean_thx_lower_ratio = np.mean(all_thx_lower_ratio) if all_thx_lower_ratio else 0
    mean_thx_rollon = np.mean(all_thx_rollon) if all_thx_rollon else 0
    mean_thx_lowpower = np.mean(all_thx_lowpower) if all_thx_lowpower else 0
    weighted_mean_lower_centroid = np.average(all_lower_centroid, weights=all_tpower)
    weighted_mean_lower_ratio = np.average(all_lower_ratio, weights=all_tpower)
    weighted_mean_rollon = np.average(all_rollon, weights=all_tpower)
    weighted_mean_lowpower = np.average(all_lowpower, weights=all_tpower)

    mean_normalised_lower_centroid = np.mean(all_normalised_lower_centroid)
    mean_normalised_lower_ratio = np.mean(all_normalised_lower_ratio)
    mean_normalised_rollon = np.mean(all_normalised_rollon)
    mean_normalised_lowpower = np.mean(all_normalised_lowpower)
    mean_normalised_thx_lower_centroid = np.mean(all_normalised_thx_lower_centroid)
    mean_normalised_thx_lower_ratio = np.mean(all_normalised_thx_lower_ratio)
    mean_normalised_thx_rollon = np.mean(all_normalised_thx_rollon)
    mean_normalised_thx_lowpower = np.mean(all_normalised_thx_lowpower)
    weighted_mean_normalised_lower_centroid = np.average(all_normalised_lower_centroid, weights=all_normalised_tpower)
    weighted_mean_normalised_lower_ratio = np.average(all_normalised_lower_ratio, weights=all_normalised_tpower)
    weighted_mean_normalised_rollon = np.average(all_normalised_rollon, weights=all_normalised_tpower)
    weighted_mean_normalised_lowpower = np.average(all_normalised_lowpower, weights=all_normalised_tpower)

    mean_centroid = np.mean(centroid)
    weighted_mean_centroid = np.average(all_centroid, weights=all_tpower)

    limited_mean_lower_centroid = np.float64(crossover_frequency) if mean_lower_centroid > crossover_frequency else mean_lower_centroid
    limited_mean_thx_lower_centroid = np.float64(crossover_frequency) if mean_thx_lower_centroid > crossover_frequency else mean_thx_lower_centroid
    limited_weighted_mean_lower_centroid = np.float64(crossover_frequency) if weighted_mean_lower_centroid > crossover_frequency else weighted_mean_lower_centroid
    limited_mean_normalised_lower_centroid = np.float64(crossover_frequency) if mean_normalised_lower_centroid > crossover_frequency else mean_normalised_lower_centroid
    limited_mean_normalised_thx_lower_centroid = np.float64(crossover_frequency) if mean_normalised_thx_lower_centroid > crossover_frequency else mean_normalised_thx_lower_centroid
    limited_weighted_mean_normalised_lower_centroid = np.float64(crossover_frequency) if weighted_mean_normalised_lower_centroid > crossover_frequency else weighted_mean_normalised_lower_centroid


    if dev_output:
        return mean_lower_centroid, mean_lower_ratio, mean_rollon, \
               mean_thx_lower_centroid, mean_thx_lower_ratio, mean_thx_rollon, \
               weighted_mean_lower_centroid, weighted_mean_lower_ratio, weighted_mean_rollon, \
               mean_lowpower, mean_thx_lowpower, weighted_mean_lowpower, \
               mean_normalised_lower_centroid, mean_normalised_lower_ratio, mean_normalised_rollon, \
               mean_normalised_lowpower, mean_normalised_thx_lower_centroid, mean_normalised_thx_lower_ratio, \
               mean_normalised_thx_rollon, mean_normalised_thx_lowpower, weighted_mean_normalised_lower_centroid, \
               weighted_mean_normalised_lower_ratio, weighted_mean_normalised_rollon, \
               weighted_mean_normalised_lowpower, mean_centroid, weighted_mean_centroid, \
               SC, flogSC, logSC, floglogSC, log_rollon, lin_rollon, \
               mean_my_duration, lp_mean_my_duration, \
               limited_mean_lower_centroid, limited_mean_thx_lower_centroid, limited_weighted_mean_lower_centroid, \
               limited_mean_normalised_lower_centroid, limited_mean_normalised_thx_lower_centroid, \
               limited_weighted_mean_normalised_lower_centroid, \
               mean_duration_ratio, mean_normalised_duration_ratio, lowband_level, weighed_lowband_level, \
               midlowband_level, weighed_midlowband_level, pitch_estimate


        # mean_lower_centroid, mean_lower_ratio, mean_rollon, depth
    else:
        '''
         Perform linear regression to obtain depth
        '''
        # coefficients from linear regression
        coefficients = np.array([-0.0370864447684, 152.569714872, -0.0577703172507, -1.83339263537, 0.000716827172454,
                                 -0.156630837052, 0.0133395675543, 10.8565562292])

        # pack metrics into a matrix
        all_metrics = np.zeros(8)

        all_metrics[0] = mean_lower_centroid
        all_metrics[1] = mean_lower_ratio
        all_metrics[2] = mean_rollon
        all_metrics[3] = np.array(mean_lower_centroid) * np.array(mean_lower_ratio)
        all_metrics[4] = np.array(mean_lower_centroid) * np.array(mean_rollon)
        all_metrics[5] = np.array(mean_lower_ratio) * np.array(mean_rollon)
        all_metrics[6] = np.array(mean_lower_centroid) * np.array(mean_lower_ratio) * np.array(mean_rollon)
        all_metrics[7] = 1.0

        # perform linear regression
        depth = np.sum(all_metrics * coefficients)

        # this return is maintained for testing in the future
        # return mean_lower_centroid, mean_lower_ratio, mean_rollon, depth

        return depth

