from __future__ import division
import numpy as np
import soundfile as sf
import timbral_util
from scipy.signal import spectrogram, hilbert
import matplotlib as mpl
import matplotlib.pyplot as plt
import librosa


def cross_covariance(x, y):
    """
      Calculates the normalised cross covariance.
    """
    x = np.array(x)
    y = np.array(y)

    product = x * y
    N = len(x)
    inv_N = 1.0 / N
    sum_x = np.sum(x)
    sum_y = np.sum(y)
    numerator = np.sum(product) - (sum_x * sum_y * inv_N)

    var_x = np.sum(x * x) - (sum_x * sum_x * inv_N)
    var_y = np.sum(y * y) - (sum_y * sum_y * inv_N)

    denominator = np.sqrt(var_x * var_y)

    return numerator / denominator


def calc_z(f):
    # frequency to bark
    z1 = 13.0 * np.arctan(0.00076 * f)
    z2 = 3.5 * np.arctan((f / 7500.0) ** 2.0)
    return z1 + z2


def calc_lower_contribution(bark, f, amplitude):
    '''
      Bark is the current bark bin, in bark
      f if the current frequency, in Hz.

    '''
    # lower slope, with a -27dB/Bark rolloff.
    mag = 20 * np.log10(amplitude)

    # how far is this in bark
    z = calc_z(f)
    delta_z = bark - z
    gain = -27 * delta_z
    return amplitude * timbral_util.db2mag(gain)


def calc_upper_contribution(bark, f, amplitude, correction=50):
    magnitude = 20 * np.log10(amplitude)

    slope = 24 + (230 / f) - (0.2 * (magnitude+correction))

    # how far is this in bark
    z = calc_z(f)
    delta_z = z - bark
    gain = -1.0 * slope * delta_z
    return amplitude * timbral_util.db2mag(gain)


def calc_fluctuation_strength(audio_samples, fs, weighted_mean=True):
    # get the FFT
    FFT_freq_step = 0.5
    NFFT = int(fs / FFT_freq_step)

    if len(audio_samples) >= NFFT+NFFT/4:
        overlap = 3 * NFFT / 4
    else:
        overlap = 0
    # get spectrogram
    freq, time, spec = spectrogram(audio_samples, fs, 'boxcar', NFFT, overlap, NFFT, 'constant', False,
                                                'density', mode='complex')
    # get absolute
    abs_spec = np.abs(spec[:int(spec.shape[0] / 2), :])

    # calculate thresholds
    thresh = timbral_util.db2mag(-60)

    bark_nos = np.arange(0.5, 24, 0.5)

    hold_critical_band_audio = np.zeros([NFFT, len(bark_nos)])
    hold_critical_band_envelope = np.zeros_like(hold_critical_band_audio)

    all_FS = []
    all_RMS = []

    for i in range(abs_spec.shape[1]):
        this_complex_spec = spec[:, i]
        this_spec = abs_spec[:, i]
        all_RMS.append(np.sum(this_spec))
        this_spec -= thresh
        this_spec = np.clip(this_spec, 0, max(this_spec))

        # for calculating the values per index
        idxs = np.where(this_spec > 0)[0]

        if len(idxs):
            m_barkband = np.zeros_like(bark_nos)

            for bark_idx, bark_no in enumerate(bark_nos):
                this_bark_filter = np.zeros_like(this_spec)
                for idx in idxs:

                    # what's this frequency and level
                    this_freq = freq[idx]
                    this_level = this_spec[idx]

                    this_bark = calc_z(this_freq)

                    if this_bark > bark_no:
                        this_bark_filter[idx] = calc_upper_contribution(bark_no, this_freq, this_level)
                    else:
                        this_bark_filter[idx] = calc_lower_contribution(bark_no, this_freq, this_level)

                # plt.plot(this_bark_filter)
                this_bark_filter_doubled = np.concatenate((this_bark_filter[:int(len(this_bark_filter))],
                                                           np.flip(this_bark_filter[:int(len(this_bark_filter))], 0)))

                to_convert_to_audio = this_complex_spec * this_bark_filter_doubled
                reconstructed_audio = np.fft.ifft(to_convert_to_audio)
                reconstructed_audio = reconstructed_audio.real
                # test_audio = np.fft.ifft(this_complex_spec)

                hold_critical_band_audio[:, bark_idx] = reconstructed_audio

                '''
                  Estimate the modulation depth
                '''
                h_mean = np.mean(np.abs(reconstructed_audio))

                # define the filter H(f_mod)
                h_BP = np.abs(reconstructed_audio.real)
                h_BP = timbral_util.filter_audio_lowpass(h_BP, 12, fs, order=4)
                h_BP = timbral_util.filter_audio_highpass(h_BP, 3.1, fs, order=2)

                h_RMS = np.sqrt(np.mean(h_BP * h_BP))

                m_i = h_RMS / h_mean

                if m_i > 0.7:
                    overload = m_i - 0.7
                    m_i = 0.7 + (overload / 3.0)

                hold_critical_band_envelope[:, bark_idx] = reconstructed_audio
                m_barkband[bark_idx] = m_i

            # try the normalised cross covariance
            k1 = np.ones(len(m_barkband))
            k2 = np.ones_like(k1)

            for j in range(len(k1)):
                # lower cross covariance
                if j > 1:
                    k1[j] = cross_covariance(hold_critical_band_envelope[j], hold_critical_band_envelope[j - 2])

                if j < len(k1) - 2:
                    k2[j] = cross_covariance(hold_critical_band_envelope[j], hold_critical_band_envelope[j + 2])

            # frequency weighting function
            g_z = np.ones_like(k1)
            g_z[np.where(bark_nos == 15)[0][0]:] = np.linspace(1, 0.5, len(bark_nos) - np.where(bark_nos == 15)[0][0])

            # variables
            p_m = 1.7
            p_k = 1.7
            p_s = 1.0

            # calculate the fluctuation strength
            FS = 0
            for j in range(len(k1)):
                FS += (m_barkband[j] ** p_m) * ((np.abs(k1[j] * k2[j])) ** p_k) * (g_z[j] ** p_s)

            # get a list of the time varying fluctuation strength
            all_FS.append(FS)
        else:
            all_FS.append(0)

    if weighted_mean:
        return np.average(all_FS, weights=all_RMS)
    else:
        return np.mean(all_FS)


def calc_critical_band_spectrum_sum(audio_samples, fs, weighted_mean=True):
    # get the FFT
    FFT_freq_step = 0.5
    NFFT = int(fs / FFT_freq_step)

    if len(audio_samples) >= NFFT+NFFT/4:
        overlap = 3 * NFFT / 4
    else:
        overlap = 0
    # get spectrogram
    freq, time, spec = spectrogram(audio_samples, fs, 'boxcar', NFFT, overlap, NFFT, 'constant', False,
                                                'density', mode='complex')
    # get absolute
    abs_spec = np.abs(spec[:int(spec.shape[0] / 2), :])

    # calculate thresholds
    thresh = timbral_util.db2mag(-60)

    bark_nos = np.arange(0.5, 24, 0.5)

    hold_critical_band_audio = np.zeros([NFFT, len(bark_nos)])
    hold_critical_band_envelope = np.zeros_like(hold_critical_band_audio)

    all_FS = []
    all_RMS = []

    for i in range(abs_spec.shape[1]):
        this_complex_spec = spec[:, i]
        this_spec = abs_spec[:, i]
        all_RMS.append(np.sum(this_spec))

        this_spec -= thresh
        this_spec = np.clip(this_spec, 0, max(this_spec))

        # for calculating the values per index
        idxs = np.where(this_spec > 0)[0]

        if len(idxs):
            m_barkband = np.zeros_like(bark_nos)

            for bark_idx, bark_no in enumerate(bark_nos):
                this_bark_filter = np.zeros_like(this_spec)
                for idx in idxs:

                    # what's this frequency and level
                    this_freq = freq[idx]
                    this_level = this_spec[idx]

                    this_bark = calc_z(this_freq)

                    if this_bark > bark_no:
                        this_bark_filter[idx] = calc_upper_contribution(bark_no, this_freq, this_level)
                    else:
                        this_bark_filter[idx] = calc_lower_contribution(bark_no, this_freq, this_level)

                # plt.plot(this_bark_filter)
                this_bark_filter_doubled = np.concatenate((this_bark_filter[:int(len(this_bark_filter))],
                                                           np.flip(this_bark_filter[:int(len(this_bark_filter))], 0)))

                to_convert_to_audio = this_complex_spec * this_bark_filter_doubled
                reconstructed_audio = np.fft.ifft(to_convert_to_audio)
                reconstructed_audio = reconstructed_audio.real
                # test_audio = np.fft.ifft(this_complex_spec)

                hold_critical_band_audio[:, bark_idx] = reconstructed_audio

                '''
                  Estimate the modulation depth
                '''
                h_mean = np.mean(np.abs(reconstructed_audio))

                # define the filter H(f_mod)
                h_BP = np.abs(reconstructed_audio.real)
                h_BP = timbral_util.filter_audio_lowpass(h_BP, 150, fs, order=4)
                h_BP = timbral_util.filter_audio_highpass(h_BP, 3.1, fs, order=2)

                h_RMS = np.sqrt(np.mean(h_BP * h_BP))

                m_i = h_RMS / h_mean

                if m_i > 0.7:
                    overload = m_i - 0.7
                    m_i = 0.7 + (overload / 3.0)

                hold_critical_band_envelope[:, bark_idx] = reconstructed_audio
                m_barkband[bark_idx] = m_i

            # try the normalised cross covariance
            k1 = np.ones(len(m_barkband))
            k2 = np.ones_like(k1)

            for j in range(len(k1)):
                # lower cross covariance
                if j > 1:
                    k1[j] = cross_covariance(hold_critical_band_envelope[j], hold_critical_band_envelope[j - 2])

                if j < len(k1) - 2:
                    k2[j] = cross_covariance(hold_critical_band_envelope[j], hold_critical_band_envelope[j + 2])

            # frequency weighting function
            g_z = np.ones_like(k1)
            g_z[np.where(bark_nos == 15)[0][0]:] = np.linspace(1, 0.5, len(bark_nos) - np.where(bark_nos == 15)[0][0])

            # variables
            p_m = 1.7
            p_k = 1.7
            p_s = 1.0

            # calculate the fluctuation strength
            FS = 0
            for j in range(len(k1)):
                FS += (m_barkband[j] ** p_m) * ((np.abs(k1[j] * k2[j])) ** p_k) * (g_z[j] ** p_s)

            # get a list of the time varying fluctuation strength
            all_FS.append(FS)
        else:
            all_FS.append(0)

    if weighted_mean:
        return np.average(all_FS, weights=all_RMS)
    else:
        return np.mean(all_FS)



def normal_dist(array, theta=1.0, mean=0.0):
    y = (1.0 / (theta * np.sqrt(2.0 * np.pi))) * np.exp((-1.0 * ((array - mean)**2.0)) / 2.0 * (theta ** 2.0))
    return y


def a_function(f):
    # return the a weighted value of the input
    t = (12194 ** 2.0) * (f ** 4.0)
    d1 = (f ** 2.0) + (20.6 ** 2.0)
    d2 = f ** 2.0 + 107.7 **2.0
    d3 = f ** 2.0 + 737.9 ** 2.0
    d4 = f ** 2.0 + 12194 ** 2.0
    W = t / (d1 * np.sqrt(d2 * d3) * d4)

    return W

def b_function(f):
    # return the a weighted value of the input
    t = (12194 ** 2.0) * (f ** 3.0)
    d1 = (f ** 2.0) + (20.6 ** 2.0)
    d2 = (f ** 2.0) + (158.5 ** 2.0)
    d3 = (f ** 2.0) + (12194 ** 2.0)
    W = t / (d1 * np.sqrt(d2) * d3)

    return W

def c_function(f):
    # return the a weighted value of the input
    t = (12194 ** 2.0) * (f ** 2.0)
    d1 = (f ** 2.0) + (20.6 ** 2.0)
    d2 = (f ** 2.0) + (12194 ** 2.0)
    W = t / (d1 * d2)

    return W


def freq_weight(f1, f2, function='a'):
    # performans a frequency weighting for each pair of frequencies.
    # roughness means more for frequencies in the upper-mids.
    means = (np.array(f1) + np.array(f2)) / 2.0
    W = np.zeros_like(f1)
    for f1idx in range(len(f1)):
        for f2idx in range(len(f2)):
            if function == 'a':
                W[f1idx, f2idx] = a_function(means[f1idx, f2idx])
            elif function == 'b':
                W[f1idx, f2idx] = b_function(means[f1idx, f2idx])
            elif function == 'c':
                W[f1idx, f2idx] = c_function(means[f1idx, f2idx])
            else:
                raise ValueError('Function type must be a, b, or c')

    return W


def plomp(f1, f2):
    b1 = 3.51
    b2 = 5.75
    xstar = 0.24
    s1 = 0.0207
    s2 = 18.96
    s = np.tril(xstar / ((s1 * np.minimum(f1, f2)) + s2))
    pd = np.exp(-b1 * s * np.abs(f2 - f1)) - np.exp(-b2 * s * np.abs(f2 - f1))
    return pd


def isodd(num):
    return num & 0x1


def matlab_hilbert(signal):
    '''
      Define a method for calculating the hilbert transform of a 1D array using the method from Matlab

    :param signal:
    :return:
    '''
    # get the fft
    n = len(signal)
    x = np.fft.fft(signal)
    h = np.zeros(n)

    if (n>0) and (~isodd(n)):
        # even and nonempty
        h[0] = 1
        h[int(n/2)] = 1
        h[1:int(n/2)] = 2
    elif n>0:
        # odd and nonempty
        h[0] = 1
        h[1:int((n+1)/2.0)] = 2

    # this is the hilbert bit
    x = np.fft.ifft(x * h)

    return x


def timbral_roughness(fname, dev_output=False, phase_correction=False):
    """
     This function is an implementation of the Vassilakis [2007] model of roughness.
     The peak picking algorithm implemented is based on the MIR toolbox's implementation.

     Version 0.2


     Vassilakis, P. 'SRA: A Aeb-based researh tool for spectral and roughness analysis of sound signals', Proceedings
     of the 4th Sound and Music Computing Conference, Lefkada, Greece, July, 2007.

    Required parameter
    :param fname:                   Audio filename to be analysed, including full file path and extension.

    Optional parameters
    :param dev_output:              Bool, when False return the roughness, when True return all extracted features
                                    (current none).
    :param phase_correction:        If the inter-channel phase should be estimated when performing a mono sum.
                                    Defaults to False.

    :return:        Roughness of the audio signal.

    Copyright 2018 Andy Pearce, Institute of Sound Recording, University of Surrey, UK.

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.
    """
    # use pysoundfile to read audio
    audio_samples, fs = sf.read(fname, always_2d=False)
    audio_samples = timbral_util.channel_reduction(audio_samples, phase_correction=phase_correction)
    audio_samples = np.lib.pad(audio_samples, (512, 0), 'constant', constant_values=(0.0, 0.0))

    # '''
    #   Fluctuation strength
    # '''
    # FS = calc_fluctuation_strength(audio_samples, fs)
    # RS = calc_critical_band_spectrum_sum(audio_samples, fs)

    # lp_audio_samples = timbral_util.filter_audio_lowpass(audio_samples, 4000, fs)
    # FS_filt = calc_fluctuation_strength(lp_audio_samples, fs)
    # RS_filt = calc_critical_band_spectrum_sum(lp_audio_samples, fs)

    # F = librosa.stft(audio_samples)
    # F = np.mean(F, axis=1)
    #
    # freq = np.linspace(1, fs/2, len(F))

    '''
      Octave band decomposition
    '''
    centre_frequencies = [500, 1000, 2000, 4000, 8000]
    envelope_spectrums = []
    weighted_envelope_spectrums = []
    for centre_frequency in centre_frequencies:
        filtered_audio = timbral_util.filter_audio_bandpass(audio_samples, centre_frequency, 1, fs)
        filtered_audio = timbral_util.filter_audio_bandpass(filtered_audio, centre_frequency, 1, fs)
        filtered_audio = timbral_util.filter_audio_bandpass(filtered_audio, centre_frequency, 1, fs)

        filtered_audio_envelope = timbral_util.fast_hilbert_spectrum(filtered_audio)
        envelope_spectrums.append(filtered_audio_envelope)
        freq = np.linspace(1, fs / 2, len(filtered_audio_envelope))
        # plt.semilogx(freq, filtered_audio_envelope)
        # plt.xlim([5, 300])
        # plt.close()

        lf_idx = np.argmax(freq>5)
        hf_idx = np.argmax(freq>300)
        mean_idx = (lf_idx + hf_idx) / 2.0
        array = np.arange(lf_idx, hf_idx)
        weight = timbral_util.normal_dist(array, theta=0.01, mean = mean_idx)
        weight -= np.min(weight)
        weight /= np.max(weight)

        array = np.zeros_like(freq)
        array[lf_idx:hf_idx] = weight

        weighted_envelope_spectrums.append(filtered_audio_envelope*array)


    sum_weighted_envelope = np.sum(weighted_envelope_spectrums, axis=1)


    # '''
    #   Get envelope and check for peak rate.
    # '''
    # envelope = timbral_util.fast_hilbert(audio_samples)
    # lp_envelope = timbral_util.filter_audio_lowpass(envelope, 150, fs, order=6)
    #
    # max_time = fs / (2 * 150) # 150Hz, *2 to half
    #
    # peaks = librosa.util.peak_pick(lp_envelope, pre_max=max_time, post_max=max_time, pre_avg=0, post_avg=1,
    #                                delta=0, wait=0)
    #
    # envelope_dynamic_range = np.max(lp_envelope) - np.min(lp_envelope)
    # thx_peaks = []
    #
    # # calculate the inter-peak dynamic range
    # inter_peak_dynamic_range = []
    # thx_inter_peak_dynamic_range = []
    # for peak_idx in range(len(peaks)):
    #     if peak_idx == len(peaks)-1:
    #         inter_peak_range = lp_envelope[peaks[peak_idx]:]
    #     else:
    #         inter_peak_range = lp_envelope[peaks[peak_idx]:peaks[peak_idx+1]]
    #
    #     dyn_range = np.max(inter_peak_range) - np.min(inter_peak_range)
    #     inter_peak_dynamic_range.append(dyn_range)
    #     if dyn_range > (envelope_dynamic_range * 0.02):
    #         thx_peaks.append(peaks[peak_idx])
    #         thx_inter_peak_dynamic_range.append(dyn_range)
    #
    #
    # # plt.plot(lp_envelope)
    # # for x in thx_peaks:
    # #     plt.axvline(x=x, color='r')
    #
    # # detected peak rate
    # diff = np.diff(thx_peaks)
    # mean_diff = np.mean(diff)
    #
    # peak_rate = mean_diff / fs
    # peak_variance = np.mean(thx_inter_peak_dynamic_range)

    # x = 1


    # try high passing the audio first?
    highpass_frequency = 700
    # audio_samples = timbral_util.filter_audio_highpass(audio_samples, highpass_frequency, fs)
    # audio_samples = timbral_util.filter_audio_highpass(audio_samples, highpass_frequency, fs)
    # audio_samples = timbral_util.filter_audio_highpass(audio_samples, highpass_frequency, fs)
    # audio_samples = timbral_util.filter_audio_highpass(audio_samples, highpass_frequency, fs)
    # audio_samples = timbral_util.filter_audio_highpass(audio_samples, highpass_frequency, fs)
    # audio_samples = timbral_util.filter_audio_highpass(audio_samples, highpass_frequency, fs)

    # try lowpassing as well?
    # lowpass_frequency = 18000
    # audio_samples = timbral_util.filter_audio_lowpass(audio_samples, lowpass_frequency, fs)
    # audio_samples = timbral_util.filter_audio_lowpass(audio_samples, lowpass_frequency, fs)
    # audio_samples = timbral_util.filter_audio_lowpass(audio_samples, lowpass_frequency, fs)
    # audio_samples = timbral_util.filter_audio_lowpass(audio_samples, lowpass_frequency, fs)
    # audio_samples = timbral_util.filter_audio_lowpass(audio_samples, lowpass_frequency, fs)
    # audio_samples = timbral_util.filter_audio_lowpass(audio_samples, lowpass_frequency, fs)



    # normalise
    # audio_samples /= np.max(np.abs(audio_samples))  # normalisation removed due to level alignment









    '''
        Change calucltion method to use spectrogram instead of this custom method
    '''
    # nfft = 4096
    # hop_size = 3 * nfft / 4
    # freq, time, spec = spectrogram(audio_samples, fs, 'hamming', nfft, hop_size, nfft, 'constant', True, 'spectrum')
    #
    # reshape audio


    '''
      Old version
    '''
    audio_len = len(audio_samples)
    time_step = 0.05
    step_samples = int(fs * time_step)
    nfft = step_samples
    window = np.hamming(nfft + 2)
    window = window[1:-1]
    olap = nfft / 2
    num_frames = int((audio_len)/(step_samples-olap))
    next_pow_2 = np.log(step_samples) / np.log(2)
    next_pow_2 = 2 ** int(next_pow_2 + 1)

    reshaped_audio = np.zeros([next_pow_2, num_frames])

    i = 0
    start_idx = int((i * (nfft / 2.0)))

    # get all the audio
    while start_idx+step_samples <= audio_len:
        audio_frame = audio_samples[start_idx:start_idx+step_samples]

        # apply window
        audio_frame = audio_frame * window

        # append zeros
        reshaped_audio[:step_samples, i] = audio_frame

        # increase the step
        i += 1
        start_idx = int((i * (nfft / 2.0)))

    spec = np.fft.fft(reshaped_audio, axis=0)
    spec_len = int(next_pow_2/2) + 1
    spec = spec[:spec_len, :]
    spec = np.absolute(spec)

    freq = fs/2 * np.linspace(0, 1, spec_len)





    # normalise
    norm_spec = (spec - np.min(spec)) / (np.max(spec) - np.min(spec))

    ''' Peak picking algorithm '''
    # cthr = 0.01  # threshold for peak picking
    cthr = 0.001  # threshold for peak picking

    _, no_segments = np.shape(spec)

    allpeakpos = []
    allpeaklevel = []
    allpeaktime = []
    all_rms = []
    for i in range(0, no_segments):
        d = norm_spec[:, i]
        d_un = spec[:, i]

        all_rms.append(np.sum(d_un))

        # find peak candidates
        peak_pos, peak_level, peak_x = timbral_util.detect_peaks(d, cthr=cthr, unprocessed_array=d_un, freq=freq)

        allpeakpos.append(peak_pos)
        allpeaklevel.append(peak_level)
        allpeaktime.append(peak_x)

    ''' get the roughness '''
    allroughness = []
    allMyRoughness = []
    maxroughness = []
    # for each frame
    for frame in range(len(allpeaklevel)):
        if frame > 150:
            stop = True
        frame_freq = allpeaktime[frame]
        frame_level = allpeaklevel[frame]

        if len(frame_freq) > 1:
            f2 = np.kron(np.ones([len(frame_freq), 1]), frame_freq)
            f1 = f2.T
            v2 = np.kron(np.ones([len(frame_level), 1]), frame_level)
            v1 = v2.T

            X = v1 * v2
            Y = (2 * v2) / (v1 + v2)
            Z = plomp(f1, f2)
            W = freq_weight(f1, f2, function='a')
            rough = (X ** 0.1) * (0.5 * (Y ** 3.11)) * Z
            myrough = rough * W

            allroughness.append(np.sum(rough))
            allMyRoughness.append(np.sum(myrough))
        else:
            allroughness.append(0)
            allMyRoughness.append(0)

    if np.max(allroughness) < 0.0001:
        maxroughness = 0.0001
        mean_roughness = 0.0001
        median_roughness = 0.0001
        mean_weighted_roughness = 0.0001

        mean_weighted_myroughness = 0.0001
        mean_myroughness = 0.0001

        capped_roughness = 0.0001
    else:
        maxroughness = np.max(allroughness)
        mean_roughness = np.mean(allroughness)
        median_roughness = np.median(allroughness)
        mean_weighted_roughness = np.average(allroughness, weights=all_rms)

        mean_weighted_myroughness = np.average(allMyRoughness, weights=all_rms)
        mean_myroughness = np.mean(allMyRoughness)

        if mean_roughness > 50:
            capped_roughness = mean_roughness
        else:
            capped_roughness = 50


    # segment audio
















    '''
      Loudness in the upper mids
    '''
    windowed_samples = timbral_util.window_audio(audio_samples)
    windowed_rms = []
    windowed_loud_spec = []
    for i in range(windowed_samples.shape[0]):
        samples = windowed_samples[i, :]
        N_entire, N_single = timbral_util.specific_loudness(samples, Pref=100.0, fs=fs, Mod=0)
        # append the loudness spec
        windowed_loud_spec.append(N_single)

        windowed_rms.append(np.sqrt(np.mean(samples * samples)))

    # need to define a function for the roughness stimuli, emphasising the 125 - 200 region (of the bark scale)
    min_bark_band = 120
    max_bark_band = 240
    mean_bark_band = (min_bark_band + max_bark_band) / 2.0
    array = np.arange(min_bark_band, max_bark_band)
    x = normal_dist(array, theta=0.01, mean=mean_bark_band)
    x -= np.min(x)
    x /= np.max(x)

    array = np.zeros(240)
    array[min_bark_band:max_bark_band] = x

    weighted_loud_spec = []
    weighted_loud_spec_sum = []
    loud_spec_centroid = []
    bark_scale = np.linspace(0.1, 24, len(windowed_loud_spec[0]))
    for i in range(len(windowed_loud_spec)):
        loud_spec = windowed_loud_spec[i]
        weighted_loud_spec.append(loud_spec * array)
        weighted_loud_spec_sum.append(np.sum(loud_spec * array))

        loud_spec_centroid.append(np.sum(loud_spec * bark_scale) / np.sum(bark_scale))




    weighted_mean_loud_spec_sum = np.average(weighted_loud_spec_sum, weights=windowed_rms)
    wegithed_mean_loud_spec_centroid = np.average(loud_spec_centroid, weights=windowed_rms)

    # overall loudspec centroid
    N_single, N_entire = timbral_util.specific_loudness(audio_samples, Pref=100.0, fs=fs, Mod=0)

    overall_loudspec_centroid = np.sum(N_entire * bark_scale) / np.sum(bark_scale)





    #############################################################################################
    '''
      Envelope spectrum
    '''
    # highpass_audio = timbral_util.filter_audio_highpass(audio_samples, 500, fs, order=6)
    # highpass_audio = timbral_util.filter_audio_highpass(highpass_audio, 500, fs)
    # highpass_audio = timbral_util.filter_audio_highpass(highpass_audio, 500, fs)

    # get analytic envelope
    # env = timbral_util.fast_hilbert(highpass_audio, use_matlab_hilbert=True)
    # env1 = np.abs(matlab_hilbert(highpass_audio))
    #
    # ENV = np.abs(np.fft.fft(env))
    # ENV = ENV[0:int(len(ENV)/2.0)]
    # env = np.abs(audio_samples)
    # filt_env = timbral_util.filter_audio_lowpass(env, 100, fs, order=6)
    # filt_env = timbral_util.filter_audio_highpass(filt_env, 10, fs, order=2)
    # windowed_env = timbral_util.window_audio(filt_env, 16384)
    # all_env_spectral_sum = []
    # all_env_rms = []
    # for i in range(windowed_env.shape[0]):
    #     samples = windowed_env[i, :]
    #     all_env_spectral_sum.append(np.sum(np.abs(samples)))
    #     all_env_rms.append(np.sqrt(np.mean(samples * samples)))
    #
    # env_spectral_sum = np.mean(all_env_spectral_sum)



    # ENV = spectrogram(env, fs)
    #
    # ENV = timbral_util.fast_hilbert_spectrum(audio_samples, use_matlab_hilbert=True)
    # env_fls = np.linspace(0, int(fs/2), len(ENV))
    # env300_idx = np.argmax(env_fls>300)
    # env15_idx = np.argmax(env_fls>15)
    # env70_idx = np.argmax(env_fls>=70)
    #
    # # array that describes the mask
    # sub_mask = np.arange(env15_idx, env300_idx)
    # # get the curve
    # sub_mask = normal_dist(np.log10(sub_mask), theta=5, mean=np.log10(env70_idx))
    # # normalise sub_mask
    # sub_mask -= np.min(sub_mask)
    # sub_mask /= np.max(sub_mask)
    # mask = np.zeros_like(env_fls)
    # mask[env15_idx:env300_idx] = sub_mask
    #
    #
    # # apply the mask
    # envelope_spectrum_masked_sum = np.sum(mask * ENV)

    if dev_output:
        return [mean_roughness, weighted_mean_loud_spec_sum, wegithed_mean_loud_spec_centroid,
                overall_loudspec_centroid, capped_roughness, maxroughness,
                mean_weighted_roughness, mean_weighted_myroughness, mean_myroughness,
                sum_weighted_envelope[0], sum_weighted_envelope[1], sum_weighted_envelope[2], sum_weighted_envelope[3],
                sum_weighted_envelope[4]]
        # return [mean_roughness, weighted_mean_loud_spec_sum, wegithed_mean_loud_spec_centroid,
        #         overall_loudspec_centroid, env_spectral_sum, capped_roughness, maxroughness,
        #         mean_weighted_roughness, mean_weighted_myroughness, mean_myroughness,
        #         sum_weighted_envelope[0], sum_weighted_envelope[1], sum_weighted_envelope[2], sum_weighted_envelope[3],
        #         peak_rate, peak_variance, (peak_rate*peak_variance), FS, RS, FS_filt, RS_filt, all_env_rms]
    else:
        return mean_roughness


