"""
  This is a python script for calculating timbral warmth.  This is calculated based on the model developed by
  Broughton 2012.

"""
import numpy as np
import soundfile as sf
from scipy.signal import spectrogram
import pyfilterbank
# import essentia
# import essentia.standard as es
import timbral_util
import scipy.stats
# import essentia.standard as es

def normal_dist(array, theta=1.0, mean=0.0):
    y = (1.0 / (theta * np.sqrt(2.0 * np.pi))) * np.exp((-1.0 * ((array - mean)**2.0)) / 2.0 * (theta ** 2.0))
    return y

def warm_region_cal(audio_samples, fs):
    #window the audio
    windowed_samples = timbral_util.window_audio(audio_samples)

    # need to define a function for the roughness stimuli, emphasising the 20 - 40 region (of the bark scale)
    min_bark_band = 10
    max_bark_band = 40
    mean_bark_band = (min_bark_band + max_bark_band) / 2.0
    array = np.arange(min_bark_band, max_bark_band)
    x = normal_dist(array, theta=0.01, mean=mean_bark_band)
    x -= np.min(x)
    x /= np.max(x)

    wr_array = np.zeros(240)
    wr_array[min_bark_band:max_bark_band] = x

    # need to define a second array emphasising the 20 - 40 region (of the bark scale)
    min_bark_band = 80
    max_bark_band = 240
    mean_bark_band = (min_bark_band + max_bark_band) / 2.0
    array = np.arange(min_bark_band, max_bark_band)
    x = normal_dist(array, theta=0.01, mean=mean_bark_band)
    x -= np.min(x)
    x /= np.max(x)

    hf_array = np.zeros(240)
    hf_array[min_bark_band:max_bark_band] = x

    windowed_loud_spec = []
    windowed_rms = []

    wr_vals = []
    hf_vals = []

    for i in range(windowed_samples.shape[0]):
        samples = windowed_samples[i, :]
        N_entire, N_single = timbral_util.specific_loudness(samples, Pref=100.0, fs=fs, Mod=0)

        # append the loudness spec
        windowed_loud_spec.append(N_single)
        windowed_rms.append(np.sqrt(np.mean(samples * samples)))

        wr_vals.append(np.sum(wr_array * N_single))
        hf_vals.append(np.sum(hf_array * N_single))



    mean_wr = np.mean(wr_vals)
    mean_hf = np.mean(hf_vals)
    weighted_wr = np.average(wr_vals, weights=windowed_rms)
    weighted_hf = np.average(hf_vals, weights=windowed_rms)

    return mean_wr, weighted_wr, mean_hf, weighted_hf










def timbral_warmth(fname, dev_output=False, phase_correction=False, max_FFT_frame_size=8192, calc_per_onset=True,
                   rolloff_ratio=0.85, max_WR = 12000, decay_slope_db=24):
    """
     This descriptor text needs to be changed

     This function calculates the apparent Warmth of an audio file.

    """
    if 'drone-cold-3' in fname:
        x  = 1

    # use pysoundfile to read audio
    audio_samples, fs = sf.read(fname, always_2d=False)

    audio_samples = timbral_util.channel_reduction(audio_samples)

    mean_wr, weighted_wr, mean_hf, weighted_hf = warm_region_cal(audio_samples, fs)

    '''
      New feature
    '''
    lowband_level, weighed_lowband_level = timbral_util.weighted_bark_level(audio_samples, fs, low_bark_band=10,
                                                                            upper_bark_band=60)

    upperband_level, weighed_upperband_level = timbral_util.weighted_bark_level(audio_samples, fs, low_bark_band=100,
                                                                            upper_bark_band=200)
    # overall levels
    N_entire, N_single = timbral_util.specific_loudness(audio_samples, Pref=100.0, fs=fs, Mod=0)
    # find the peak
    peak_idx = np.argmax(N_single)
    mask_len = len(N_single) - peak_idx
    mask = np.ones(mask_len)
    for i in range(mask_len):
        mask[i] -= 0.01 * i

    mask += np.max(N_single) - 1
    vals = N_single[peak_idx:]
    overmask = vals - mask
    overmask_sum = np.sum(overmask.clip(0, max(overmask)))
















    # normalise audio
    # audio_samples /= max(abs(audio_samples))

    if calc_per_onset:
        # segment the signal based on the onsetdectection algorithm
        # calculate the envelope of the signal
        envelope = timbral_util.sample_and_hold_envelope_calculation(audio_samples, fs, decay_time=0.1)
        envelope_time = np.arange(len(envelope)) / float(fs)

        # calculate the onsets
        nperseg = 4096
        original_onsets = timbral_util.calculate_onsets(audio_samples, envelope, fs, nperseg=nperseg)
    else:
        original_onsets = [0]

    # set defaults for holding
    frame_warmth = []
    frame_rms = []
    all_fundamental = []
    all_ratio = []
    all_SC = []
    all_WR_Ratio = []
    all_rolloff = []
    fundamental_to_SC = []
    WR_to_SC = []
    fundamental_to_rolloff = []
    WR_to_rolloff = []
    SC_to_rolloff = []
    all_ratio260 = []
    all_upper_sum = []
    all_lower_sum = []
    all_hp_ratio = []
    all_gradient = []
    all_decay_score = []
    all_spread = []

    # all estimates to be calculated over duration of segment, as Mark calculated
    for idx, onset in enumerate(original_onsets):
        if onset == original_onsets[-1]:
            # this is the last onset
            segment = audio_samples[onset:]
        else:
            segment = audio_samples[onset:original_onsets[idx+1]]

        segment_rms = np.sqrt(np.mean(segment * segment))
        frame_rms.append(segment_rms)

        # estimate of fundamental frequency
        # get FFT of signal
        segment_length = len(segment)
        if segment_length < max_FFT_frame_size:
            freq, time, spec = spectrogram(segment, fs, nperseg=segment_length, nfft=max_FFT_frame_size)
        else:
            freq, time, spec = spectrogram(segment, fs, nperseg=max_FFT_frame_size, nfft=max_FFT_frame_size)
            if spec.shape[1] > 1:
                spec = np.sum(spec, axis=1)
                spec = spec.flatten()

        # normalise for this onset
        spec = np.array(list(spec)).flatten()
        this_shape = spec.shape
        spec /= max(abs(spec))

        # old method
        # spec = np.abs(np.fft.fft(segment))
        # spec = spec[:int(len(spec)/2)]
        # spec /= max(abs(spec))
        #
        # freq = np.linspace(0, fs/2.0, len(spec))



        # peak picking algorithm
        peak_idx, peak_value, peak_x = timbral_util.detect_peaks(spec, freq=freq, fs=fs)
        # find lowest peak
        fundamental = np.min(peak_x)
        fundamental_idx = np.min(peak_idx)

        all_fundamental.append(fundamental)

        '''
          Warmth region calculation
        '''
        # estimate the Warmth region
        WR_upper_f_limit = fundamental * 3.5
        if WR_upper_f_limit > max_WR:
            WR_upper_f_limit = 12000
        tpower = np.sum(spec)

        WR_upper_f_limit_idx = int(np.where(freq>WR_upper_f_limit)[0][0])
        # WR_energy = np.sum(spec[fundamental_idx:WR_upper_f_limit_idx])
        WR_energy = np.sum(spec[:WR_upper_f_limit_idx])
        WR_ratio = WR_energy / float(tpower)
        all_WR_Ratio.append(WR_ratio)

        '''
          Spectral rolloff
        '''
        tpower_above_fundamental = np.sum(spec[fundamental_idx:])
        thresh = rolloff_ratio * tpower_above_fundamental
        spec_idx = len(spec) - 1
        hold_tpower = tpower_above_fundamental
        while hold_tpower > thresh:
            spec_idx -= 1
            hold_tpower = np.sum(spec[fundamental_idx:spec_idx])
        rolloff = freq[spec_idx]
        all_rolloff.append(rolloff)



        # estimate of fundamental-to-260Hz :  overall energy
        # need to check if the fundamental is greater than the
        if fundamental < 260:
            # find frequency bin closest to 260Hz
            top_level_idx = int(np.where(freq>260)[0][0])
            # sum energy up to this bin
            low_energy = np.sum(spec[fundamental_idx:top_level_idx])
            # sum all energy
            tpower = np.sum(spec)
            # take ratio
            ratio = low_energy / float(tpower)
        else:
            # make exception where fundamental is greater than
            ratio = 0

        all_ratio.append(ratio)
        # spectral centroid
        top = np.sum(freq * spec)
        bottom = float(np.sum(spec))
        SC = np.sum(freq * spec) / float(np.sum(spec))
        if SC > 10000:
            x = 1
        all_SC.append(SC)

        # apply warmth equation
        frame_warmth.append(40.721 - (0.04 * fundamental) + (0.541 * ratio) + (0.005 * SC))

        '''
          Difference metrics
        '''
        if SC <= fundamental:
            fundamental_to_SC.append(0)
        else:
            fundamental_to_SC.append(np.log10(SC - fundamental))
        if SC <= 3.5*fundamental:
            WR_to_SC.append(0)
        else:
            WR_to_SC.append(np.log10(SC - (3.5*fundamental)))
        if rolloff <= fundamental:
            fundamental_to_rolloff.append(0)
        else:
            fundamental_to_rolloff.append(np.log10(rolloff - fundamental))
        if rolloff <= 3.5 * fundamental:
            WR_to_rolloff.append(0)
        else:
            WR_to_rolloff.append(np.log10(rolloff - (3.5 * fundamental)))
        if SC >= rolloff:
            SC_to_rolloff.append(0)
        else:
            SC_to_rolloff.append(np.log10(rolloff - SC))


        '''
          260Hz ratio
        '''
        loc260 = np.where(freq>260)[0][0]
        low_tpower = np.sum(spec[:loc260])
        hi_tpower = np.sum(spec)
        ratio260 = low_tpower / float(hi_tpower)
        all_ratio260.append(ratio260)


        '''
          Bandwidth estimation
          - use spectral spread as a simple measure
        '''
        all_spread.append(np.sqrt(np.sum(((freq - SC) ** 2) * spec) / np.sum(spec)))



        '''
          Continuous/onsetoffset richness
        '''
        import librosa
        D = librosa.core.stft(audio_samples) # get librosa version of spectrogram
        H, P = librosa.decompose.hpss(D)  # separate into spectrograms
        # H, P = librosa.decompose.hpss(D, margin=(1.0, 5.0))
        percussive_audio = librosa.core.istft(P)  # invert STFT to audio
        harmonic_audio = librosa.core.istft(H)  # invert STFT to audio

        # frame by frame RMS energy
        percussive_energy = timbral_util.calculate_rms_enveope(percussive_audio, step_size=1024, overlap_step=512, normalise=False)
        harmonic_energy = timbral_util.calculate_rms_enveope(harmonic_audio, step_size=1024, overlap_step=512, normalise=False)
        hp_ratio = []
        hp_t_power = []
        for i in range(len(percussive_energy)):
            if percussive_energy[i] != 0 and harmonic_energy[0] != 0:
                hp_ratio.append(percussive_energy[i] / (percussive_energy[i] + harmonic_energy[i]))
                hp_t_power.append((percussive_energy[i] + harmonic_energy[i]))

        all_hp_ratio.append(np.average(hp_ratio, weights=hp_t_power))
        # return ratio


        '''
          HF decay
          - linear regression of the values above WR
        '''
        above_WR_spec = np.log10(spec[WR_upper_f_limit_idx:])
        above_WR_freq = np.log10(freq[WR_upper_f_limit_idx:])
        np.ones_like(above_WR_freq)
        metrics = np.array([above_WR_freq, np.ones_like(above_WR_freq)])
        from sklearn import linear_model
        model = linear_model.LinearRegression(fit_intercept=False)
        # model.fit()
        model.fit(metrics.transpose(), above_WR_spec)
        decay_score = model.score(metrics.transpose(), above_WR_spec)

        all_gradient.append(model.coef_[0])
        all_decay_score.append(decay_score)



        '''
          Ideal warmth curve
        '''
        # sum_spec = np.sum(spec, axis=1)
        # sum_spec /= max(sum_spec)
        # only consider values above the fundamental
        ideal_curve = np.ones_like(spec) * spec[fundamental_idx] # scale to fundamental
        decay_length = len(spec[WR_upper_f_limit_idx:])

        delta_y = np.log10(decay_slope_db) - np.log10(2*decay_slope_db)
        delta_x = np.log10(200) - np.log10(100)
        m = delta_y / delta_x
        decay_array = (np.linspace(freq[WR_upper_f_limit_idx], freq[-1], decay_length)) ** m
        decay_array /= decay_array[0]  # normalise it
        decay_array *= spec[fundamental_idx]  # make first element match
        ideal_curve[WR_upper_f_limit_idx:] = decay_array

        db_spec = 20*np.log10(spec)
        db_ideal =  20*np.log10(ideal_curve)

        diff_spec = db_spec - db_ideal

        upper_clip_diff_spec = diff_spec.clip(0)  # all values above
        lower_clip_diff_spec = diff_spec.clip(min(diff_spec), 0)  # all values below the ideal curve

        all_upper_sum.append(np.mean(upper_clip_diff_spec[WR_upper_f_limit_idx:]))
        all_lower_sum.append(np.mean(np.abs(lower_clip_diff_spec[fundamental_idx:WR_upper_f_limit_idx])))  # take absolue since values are negative

        # plt.semilogx(freq, db_spec)
        # plt.semilogx(freq, db_ideal)
        # plt.axvline(x=freq[fundamental_idx], color='r')
        # plt.axvline(x=freq[WR_upper_f_limit_idx], color='r')
        #
        # plt.xlim([20, fs/2])
        # plt.show()

    '''
      Essentia metrics
    '''
    # audio, fs, num_chans, checksum, bit_rate, codec = es.AudioLoader(filename=fname)()
    # if num_chans == 1:
    #     audio = audio[:, 0]
    # else:
    #     audio = np.sum(audio, axis=1)
    #
    # # initialise all essentia things needed
    # w = es.Windowing(type='hann')
    # fft = es.FFT()
    # spectrum = es.Spectrum()
    # harmonicPeaks = es.HarmonicPeaks()  # requires frequencies (spectral peaks), mangitudes (spectral peaks),
    # # pitch (fundamental frequency estimation)
    # spectral_Peaks = es.SpectralPeaks(sampleRate=fs, minFrequency=20)  # requires spectrum
    # pitch = es.PitchYin(maxFrequency=8000)  # take in time domain audio signal
    # PitchSalience = es.PitchSalience()
    # Dissonance = es.Dissonance()
    # Inharmonicity = es.Inharmonicity()
    #
    # # store variables
    # hold_dissonance = []
    # hold_pitch_salience = []
    # hold_pitch = []
    # hold_inharmonicity = []
    # all_dissonance = []
    # all_inharmonicity = []
    # all_pitch_salience = []
    #
    # for frame in es.FrameGenerator(audio, frameSize=1024, hopSize=512):
    #     freq, mags, = spectral_Peaks(spectrum(w(frame)))
    #     dissonance = Dissonance(freq, mags)
    #     hold_dissonance.append(dissonance)
    #     pitch_salience = PitchSalience(spectrum(w(frame)))
    #     hold_pitch_salience.append(pitch_salience)
    #     p, conf = pitch(frame)  # pitch (Hz), confidence (0, 1)
    #     hold_pitch.append(p)
    #     if conf > 0:  # if pitchConfidence == 0, then pitch is not defined, not to be used
    #         harm_freq, harm_mags, = harmonicPeaks(freq, mags, p)
    #         inharmonicity = Inharmonicity(harm_freq, harm_mags)
    #         hold_inharmonicity.append(inharmonicity)
    #     else:
    #         hold_inharmonicity.append(np.nan)
    #
    # all_dissonance.append(np.nanmean(hold_dissonance))
    # if np.all(np.isnan(hold_inharmonicity)):
    #     all_inharmonicity.append(1.0)
    # else:
    #     all_inharmonicity.append(np.nanmean(hold_inharmonicity))
    # all_pitch_salience.append(np.nanmean(hold_pitch_salience))









    if dev_output:
        mean_fundamental = np.log10(np.mean(all_fundamental))
        mean_ratio = np.mean(all_ratio)
        mean_SC = np.log10(np.mean(all_SC))
        mean_WR = np.mean(all_WR_Ratio)
        mean_rolloff = np.mean(all_rolloff)
        # difference metrics
        mean_fundamental_to_SC = np.mean(fundamental_to_SC)
        mean_WR_to_SC = np.mean(WR_to_SC)
        mean_fundamental_to_rolloff = np.mean(fundamental_to_rolloff)
        mean_WR_to_rolloff = np.mean(WR_to_rolloff)
        mean_SC_to_rolloff = np.mean(SC_to_rolloff)

        weighted_mean_fundamental = np.log10(np.average(all_fundamental, weights=frame_rms))
        weighted_mean_ratio = np.average(all_ratio, weights=frame_rms)
        weighted_mean_SC = np.log10(np.average(all_SC, weights=frame_rms))
        weighted_mean_WR = np.average(all_WR_Ratio, weights=frame_rms)
        # difference metrics
        weighted_mean_fundamental_to_SC = np.average(fundamental_to_SC, weights=frame_rms)
        weighted_mean_WR_to_SC = np.average(WR_to_SC, weights=frame_rms)
        weighted_mean_fundamental_to_rolloff = np.average(fundamental_to_rolloff, weights=frame_rms)
        weighted_mean_WR_to_rolloff = np.average(WR_to_rolloff, weights=frame_rms)

        # mean_dissonance = all_dissonance #np.mean(hold_dissonance)
        # mean_pitch_salience = all_pitch_salience # np.mean(hold_pitch_salience)
        # pitch = np.mean(hold_pitch)
        # mean_inharmonicity = all_inharmonicity  #np.mean(hold_inharmonicity)

        mean_ratio260 = np.mean(all_ratio260)

        mean_upper_sum = np.mean(all_upper_sum)
        mean_lower_sum = np.mean(all_lower_sum)

        mean_gradient = np.mean(all_gradient)
        mean_hp_ratio = np.mean(all_hp_ratio)
        mean_decay_score = np.mean(all_decay_score)
        mean_spread = np.mean(all_spread)

        mean_warmth = np.mean(frame_warmth)
        weighted_mean_warmth = np.average(frame_warmth, weights=frame_rms)


        return mean_fundamental, mean_ratio, mean_SC, weighted_mean_fundamental, weighted_mean_ratio, \
               weighted_mean_SC, mean_WR, weighted_mean_WR, np.log10(mean_rolloff), \
               mean_fundamental_to_SC, mean_WR_to_SC, mean_fundamental_to_rolloff, mean_WR_to_rolloff, \
               weighted_mean_fundamental_to_SC, weighted_mean_WR_to_SC, weighted_mean_fundamental_to_rolloff, \
               weighted_mean_WR_to_rolloff, \
               mean_SC_to_rolloff, mean_ratio260, mean_upper_sum, mean_lower_sum, mean_gradient, mean_hp_ratio, \
               mean_decay_score, mean_spread, mean_warmth, weighted_mean_warmth, \
               mean_wr, weighted_wr, mean_hf, weighted_hf, overmask_sum


    else:
        mean_warmth = np.mean(frame_warmth)
        weighted_mean_warmth = np.average(frame_warmth, weights=frame_rms)

        return mean_warmth, weighted_mean_warmth

        # warmth = 40.721 - (0.04 * fundamental) + (0.541 * ratio) + (0.005 * SC)